# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
import logging

import roman


def ascii_report_from_ranked_vaccine_peptides(
        ranked_variants_with_vaccine_peptides,
        mhc_alleles,
        variants,
        bam_path):
    # create dictionary mapping variants to coding effects
    variants_to_top_coding_effect_dict = {
        variant: effect_collection.top_priority_effect()
        for (variant, effect_collection)
        in variants.effects().drop_silent_and_noncoding().groupby_variant().items()
    }
    lines = [
        "VCF (somatic variants) path: %s" % variants.path,
        "BAM (RNAseq reads) path: %s" % bam_path,
        "MHC alleles: %s" % (" ".join(mhc_alleles)),
        "Total number of somatic variants: %d" % (len(variants),),
        "Somatic variants with predicted coding effects: %d" % (
            len(variants_to_top_coding_effect_dict),),
        "---",
    ]

    for i, (variant, vaccine_peptides) in enumerate(
            ranked_variants_with_vaccine_peptides):
        variant_short_description = variant.short_description
        if len(vaccine_peptides) == 0:
            logging.info(
                "Skipping %s, no vaccine peptides" % variant_short_description)
            continue

        gene_name = vaccine_peptides[0].mutant_protein_fragment.gene_name
        lines.append("\n%d) %s (%s)" % (
            i + 1,
            variant_short_description,
            gene_name,))
        lines.append(
            "\tTop score: %0.2f" % (vaccine_peptides[0].combined_score))
        lines.append(
            "\tPredicted effect: %s" % (
                variants_to_top_coding_effect_dict.get(variant)))
        lines.append(
            "\tReads supporting variant allele: %d" % (
                vaccine_peptides[0].mutant_protein_fragment.n_alt_reads))
        lines.append(
            "\tReads supporting reference allele: %d" % (
                vaccine_peptides[0].mutant_protein_fragment.n_ref_reads))
        lines.append(
            "\tReads supporting other alleles: %d" % (
                vaccine_peptides[0].mutant_protein_fragment.n_other_reads))

        lines.append("\tVaccine Peptides:")
        for j, vaccine_peptide in enumerate(vaccine_peptides):
            mutant_protein_fragment = vaccine_peptide.mutant_protein_fragment
            amino_acids = mutant_protein_fragment.amino_acids
            mutation_start = mutant_protein_fragment.mutant_amino_acid_start_offset
            mutation_end = mutant_protein_fragment.mutant_amino_acid_end_offset
            aa_before_mutation = amino_acids[:mutation_start]
            aa_mutant = amino_acids[mutation_start:mutation_end]
            aa_after_mutation = amino_acids[mutation_end:]
            lines.append("\t\t%s. %s_%s_%s (score = %0.2f)" % (
                roman.toRoman(j + 1).lower(),
                aa_before_mutation,
                aa_mutant,
                aa_after_mutation,
                vaccine_peptide.combined_score))
            lines.append(
                "\t\t   - Length: %d" % len(amino_acids))
            lines.append(
                "\t\t   - Expression score: %0.2f" % (vaccine_peptide.expression_score))
            lines.append(
                "\t\t   - Mutant epitope score: %0.2f" % (
                    vaccine_peptide.mutant_epitope_score))
            lines.append(
                "\t\t   - Wildtype epitope score: %0.2f" % (
                    vaccine_peptide.wildtype_epitope_score))

            lines.append(
                "\t\t   - Reads fully spanning cDNA sequence(s): %d" % (
                    mutant_protein_fragment.n_alt_reads_supporting_protein_sequence))
            lines.append(
                "\t\t   - Mutant amino acids: %d" % (
                    mutant_protein_fragment.n_mutant_amino_acids))
            lines.append(
                "\t\t   - Mutation distance from edge: %d" % (
                    mutant_protein_fragment.mutation_distance_from_edge))
            lines.append("\t\t   - Predicted mutant epitopes:")
            for epitope_prediction in vaccine_peptide.epitope_predictions:
                if epitope_prediction.overlaps_mutation and epitope_prediction.ic50 <= 2000:
                    lines.append("\t\t\t * %s %0.2f (%s)" % (
                        epitope_prediction.peptide_sequence,
                        epitope_prediction.ic50,
                        epitope_prediction.allele))
    return "\n".join(lines)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
import logging
from collections import OrderedDict

import pandas as pd
from isovar.protein_sequences import (
    reads_generator_to_protein_sequences_generator,
)

from .mutant_protein_fragment import MutantProteinFragment
from .epitope_prediction import predict_epitopes, slice_epitope_predictions
from .vaccine_peptide import VaccinePeptide

def vaccine_peptides_for_variant(
        variant,
        isovar_protein_sequences,
        mhc_predictor,
        vaccine_peptide_length,
        padding_around_mutation,
        max_vaccine_peptides_per_variant):
    """
    Returns list containing (MutantProteinFragment, VaccinePeptideMetrics) pairs
    """
    isovar_protein_sequences = list(isovar_protein_sequences)
    if len(isovar_protein_sequences) == 0:
        logging.info("No protein sequences for %s" % (variant,))
        return []

    protein_fragment = MutantProteinFragment.from_isovar_protein_sequence(
        variant=variant,
        protein_sequence=isovar_protein_sequences[0])

    logging.info("Mutant protein fragment for %s: %s" % (
        variant,
        protein_fragment))

    epitope_predictions = predict_epitopes(
        mhc_predictor=mhc_predictor,
        protein_fragment=protein_fragment)

    candidate_vaccine_peptides = []

    for offset, candidate_fragment in protein_fragment.top_k_subsequences(
            subsequence_length=vaccine_peptide_length,
            k=2 * padding_around_mutation + 1):
        subsequence_epitope_predictions = slice_epitope_predictions(
            epitope_predictions,
            start_offset=offset,
            end_offset=offset + len(candidate_fragment))

        assert all(
            p.source_sequence == candidate_fragment.amino_acids
            for p in subsequence_epitope_predictions)

        candidate_vaccine_peptide = VaccinePeptide(
            mutant_protein_fragment=candidate_fragment,
            epitope_predictions=subsequence_epitope_predictions)

        logging.info("%s, combined score: %0.4f" % (
            candidate_vaccine_peptide,
            candidate_vaccine_peptide.combined_score))
        candidate_vaccine_peptides.append(candidate_vaccine_peptide)

    max_score = max(vp.combined_score for vp in candidate_vaccine_peptides)
    n_total_candidates = len(candidate_vaccine_peptides)

    # only keep candidate vaccines that are within 1% of the maximum
    # combined score
    filtered_candidate_vaccine_peptides = [
        vp
        for vp in candidate_vaccine_peptides
        if vp.combined_score / max_score > 0.99
    ]

    print("Keeping %d/%d vaccine peptides for %s" % (
        len(filtered_candidate_vaccine_peptides),
        n_total_candidates,
        variant))
    filtered_candidate_vaccine_peptides.sort(key=VaccinePeptide.lexicographic_sort_key)

    if len(filtered_candidate_vaccine_peptides) > 0:
        print("\n\nTop vaccine peptides for %s:" % variant)
        for i, vaccine_peptide in enumerate(filtered_candidate_vaccine_peptides):
            print("%d) %s (combined score = %0.4f)\n" % (
                i + 1,
                vaccine_peptide,
                vaccine_peptide.combined_score))
    return filtered_candidate_vaccine_peptides[:max_vaccine_peptides_per_variant]

def generate_vaccine_peptides(
        reads_generator,
        mhc_predictor,
        vaccine_peptide_length,
        padding_around_mutation,
        max_vaccine_peptides_per_variant,
        min_reads_supporting_cdna_sequence):
    """
    Returns dictionary mapping each variant to list of VaccinePeptide objects.
    """

    # total number of amino acids is the vaccine peptide length plus the
    # number of off-center windows around the mutation
    protein_fragment_sequence_length = (
        vaccine_peptide_length + 2 * padding_around_mutation)

    protein_sequences_generator = reads_generator_to_protein_sequences_generator(
        reads_generator,
        transcript_id_whitelist=None,
        protein_sequence_length=protein_fragment_sequence_length,
        min_reads_supporting_cdna_sequence=min_reads_supporting_cdna_sequence,
        max_protein_sequences_per_variant=1)

    result_dict = {}
    for variant, isovar_protein_sequences in protein_sequences_generator:
        vaccine_peptides = vaccine_peptides_for_variant(
            variant=variant,
            isovar_protein_sequences=isovar_protein_sequences,
            mhc_predictor=mhc_predictor,
            vaccine_peptide_length=vaccine_peptide_length,
            padding_around_mutation=padding_around_mutation,
            max_vaccine_peptides_per_variant=max_vaccine_peptides_per_variant)
        result_dict[variant] = vaccine_peptides
    return result_dict

def ranked_vaccine_peptides(
        reads_generator,
        mhc_predictor,
        vaccine_peptide_length,
        padding_around_mutation,
        max_vaccine_peptides_per_variant,
        min_reads_supporting_cdna_sequence):
    """
    Returns sorted list whose first element is a Variant and whose second
    element is a list of VaccinePeptide objects.
    """
    variants_to_vaccine_peptides_dict = generate_vaccine_peptides(
        reads_generator=reads_generator,
        mhc_predictor=mhc_predictor,
        vaccine_peptide_length=vaccine_peptide_length,
        padding_around_mutation=padding_around_mutation,
        max_vaccine_peptides_per_variant=max_vaccine_peptides_per_variant,
        min_reads_supporting_cdna_sequence=min_reads_supporting_cdna_sequence)
    result_list = list(variants_to_vaccine_peptides_dict.items())
    result_list.sort(
        key=lambda x: x[1][0].combined_score if len(x[1]) > 0 else 0.0,
        reverse=True)
    for variant, vaccine_peptides in result_list:
        print("---")
        print(variant)

        for vaccine_peptide in vaccine_peptides:
            print(vaccine_peptide.mutant_protein_fragment.gene_name)
            print(vaccine_peptide.mutant_protein_fragment.amino_acids)
            print(vaccine_peptide.combined_score)
    return result_list


def dataframe_from_ranked_list(ranked_list_of_variants_with_vaccine_peptides):
    """
    Converts list of (Variant, [VaccinePeptide]) into a DataFrame, dropping
    any variants without vaccine peptide predictions.
    """
    columns = OrderedDict([
        ("genome", []),
        ("chr", []),
        ("pos", []),
        ("ref", []),
        ("alt", []),

    ])
    for field in MutantProteinFragment._fields:
        columns[field] = []
    columns["variant_rank"] = []
    columns["peptide_secondary_rank"] = []
    columns["mutant_epitope_score"] = []
    columns["wildtype_epitope_score"] = []
    columns["expression_score"] = []
    columns["combined_score"] = []

    for i, (variant, vaccine_peptides) in enumerate(
            ranked_list_of_variants_with_vaccine_peptides):
        for j, vaccine_peptide in enumerate(vaccine_peptides):
            columns["genome"].append(variant.reference_name)
            columns["chr"].append(variant.contig)
            columns["pos"].append(variant.original_start)
            columns["ref"].append(variant.original_ref)
            columns["alt"].append(variant.original_alt)

            for field in MutantProteinFragment._fields:
                columns[field].append(getattr(
                    vaccine_peptide.mutant_protein_fragment, field))
            columns["variant_rank"].append(i + 1)
            columns["peptide_secondary_rank"].append(j + 1)
            columns["mutant_epitope_score"].append(vaccine_peptide.mutant_epitope_score)
            columns["wildtype_epitope_score"].append(vaccine_peptide.wildtype_epitope_score)
            columns["expression_score"].append(vaccine_peptide.expression_score)
            columns["combined_score"].append(vaccine_peptide.combined_score)
    print(columns)
    return pd.DataFrame(columns, columns=columns.keys())



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
import sys
import logging

from isovar.args.variant_sequences import make_variant_sequences_arg_parser
from isovar.args.rna_reads import allele_reads_generator_from_args
from isovar.args.variants import variants_from_args

from topiary.commandline_args.mhc import (
    add_mhc_args,
    mhc_alleles_from_args,
    mhc_binding_predictor_from_args,
)

from .core_logic import ranked_vaccine_peptides, dataframe_from_ranked_list
from .report import ascii_report_from_ranked_vaccine_peptides

# inherit all commandline options from Isovar
arg_parser = make_variant_sequences_arg_parser(
    prog="vaxrank",
    description=(
        "Select personalized vaccine peptides from cancer variants, "
        "expression data, and patient HLA type."),
)

add_mhc_args(arg_parser)

arg_parser.add_argument(
    "--output-csv",
    default="vaccine-peptides.csv",
    help="Name of CSV file which contains predicted sequences")

arg_parser.add_argument(
    "--output-ascii-report",
    default="vaccine-peptides-report.txt",
    help="Path to ASCII vaccine peptide report")

vaccine_peptide_group = arg_parser.add_argument_group("Vaccine peptide options")
vaccine_peptide_group.add_argument(
    "--vaccine-peptide-length",
    default=25,
    type=int,
    help="Number of amino acids in the vaccine peptides (default %(default)s)")

vaccine_peptide_group.add_argument(
    "--padding-around-mutation",
    default=0,
    type=int,
    help=(
        "Number of off-center windows around the mutation to consider "
        "as vaccine peptides (default %(default)s)"
    ))

vaccine_peptide_group.add_argument(
    "--max-vaccine-peptides-per-mutation",
    default=1,
    type=int,
    help="Number of vaccine peptides to generate for each mutation")

vaccine_peptide_group.add_argument(
    "--max-mutations-in-report",
    default=10,
    type=int,
    help="Number of mutations to report")

def main(args_list=None):
    """
    Script to generate vaccine peptide predictions from somatic cancer variants,
    patient HLA type, and tumor RNA-seq data.

    Example usage:
        vaxrank
            --vcf somatic.vcf \
            --bam rnaseq.bam \
            --vaccine-peptide-length 25 \
            --output-csv vaccine-peptides.csv
    """
    if args_list is None:
        args_list = sys.argv[1:]

    logging.basicConfig(level=logging.DEBUG)
    args = arg_parser.parse_args(args_list)
    print(args)

    variants = variants_from_args(args)
    print(variants)

    mhc_alleles = mhc_alleles_from_args(args)
    print("MHC alleles: %s" % (mhc_alleles,))
    mhc_predictor = mhc_binding_predictor_from_args(args)

    # generator that for each variant gathers all RNA reads, both those
    # supporting the variant and reference alleles
    reads_generator = allele_reads_generator_from_args(args)

    ranked_list = ranked_vaccine_peptides(
        reads_generator=reads_generator,
        mhc_predictor=mhc_predictor,
        vaccine_peptide_length=args.vaccine_peptide_length,
        padding_around_mutation=args.padding_around_mutation,
        max_vaccine_peptides_per_variant=args.max_vaccine_peptides_per_mutation,
        min_reads_supporting_cdna_sequence=args.min_reads_supporting_variant_sequence)

    df = dataframe_from_ranked_list(ranked_list)
    print(df)
    df.to_csv(args.output_csv, index=False)

    ascii_report = ascii_report_from_ranked_vaccine_peptides(
        ranked_variants_with_vaccine_peptides=ranked_list,
        mhc_alleles=mhc_alleles,
        variants=variants,
        bam_path=args.bam)

    print(ascii_report)

    with open(args.output_ascii_report, "w") as f:
        f.write(ascii_report)
